#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The results of inspecting a module are as follows:

Constants
---------

All constants identified within the code shall be registered.

Classes
-------

All global classes shall be registered; local classes (within functions) or
nested classes (within classes) are not currently registered.

Base classes must be detected and constant.

All classes without bases are made to inherit from __builtins__.object in order
to support some standard methods.

Functions
---------

All functions and lambda definitions shall be registered.

Namespaces
----------

Modules define their own "global" namespace, within which classes, functions
and lambda definitions establish a hierarchy of namespaces.

Only local, global and built-in namespaces are recognised; closures are not
supported.

Assignments
-----------

Name assignment and attribute assignment involving modules and classes cause
names to be associated with values within namespaces.

Any assignments within loops are considered to cause the targets of such
assignments to provide non-constant values.

Assignments to names are only really considered to cause the targets of such
assignments to provide constant values if the targets reside in the module
namespace or in class namespaces, subject to the above conditions.

Assignments to names within functions are not generally considered to cause the
targets of such assignments to provide constant values since functions can be
invoked many times with different inputs. However, there may be benefits in
considering a local to be constant within a single invocation.
"""

from micropython.common import *
from micropython.data import *
import compiler.ast
from compiler.visitor import ASTVisitor

# Program visitors.

class InspectedModule(ASTVisitor, Module):

    """
    An inspected module, providing core details via the Module superclass, but
    capable of being used as an AST visitor.
    """

    def __init__(self, name, importer):

        """
        Initialise this visitor with a module 'name' and an 'importer' which is
        used to provide access to other modules when required.
        """

        ASTVisitor.__init__(self)
        Module.__init__(self, name)
        self.visitor = self

        # Import machinery links.

        self.importer = importer
        self.optimisations = importer.optimisations
        self.builtins = self.importer.modules.get("__builtins__")
        self.loaded = 0

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_init = 0        # Find instance attributes in __init__ methods.
        self.in_method = 0      # Find instance attributes in all methods.
        self.in_loop = 0        # Note loop "membership", affecting assignments.
        self.namespaces = []
        self.module = None

    def parse(self, filename):

        "Parse the file having the given 'filename'."

        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):

        "Process the given 'module'."

        self.astnode = self.module = module
        processed = self.dispatch(module)
        if self.has_key("__all__"):
            all = self["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self.store(n.value, self.importer.add_module(self.name + "." + n.value))
        return processed

    def vacuum(self):

        """
        Vacuum the module namespace, removing unreferenced objects and unused
        names.
        """

        for name, value in self.items():

            if self.should_optimise_unused_objects():

                # Remove unused names.

                if name not in self.importer.names_used:
                    del self[name]

                # Remove entries for unreferenced objects.
                # This, due to the nature of the referenced attribute, assumes
                # that only explicitly mentioned classes and functions are
                # employed in the final program.

                elif isinstance(value, Attr):
                    attr_value = value.value

                    # Only remove entries for classes and functions, not methods.

                    for attr_value in value.assignment_values:
                        if (isinstance(attr_value, Function) and not attr_value.is_method() or
                            isinstance(attr_value, Class)) and not attr_value.referenced:
                            pass
                        else:
                            break
                    else:
                        del self[name]

            # Complain about globals not initialised at the module level.

            if isinstance(value, Global):
                print "Warning: global %r in module %r not initialised at the module level." % (name, self.name)

        # Remove unreferenced objects.

        if self.should_optimise_unused_objects():

            all_objects = list(self.all_objects)

            for obj in all_objects:

                # Only remove entries for classes and functions, not methods.

                if (isinstance(obj, Function) and not obj.is_method() or
                    isinstance(obj, Class)) and not obj.referenced:

                    self.all_objects.remove(obj)

                # Remove unused entries from classes plus associated methods.

                if isinstance(obj, Class):
                    for name, attr in obj.class_attributes().items():

                        if name not in self.importer.names_used:
                            del obj[name]

                            # Methods can only be deleted if they are the only
                            # assigned object to the class and are unreferenced.

                            if attr.assignments == 1 and isinstance(attr.value, Function) and \
                                attr.value.is_method() and not attr.value.referenced:

                                self.all_objects.remove(attr.value)

    def finalise(self):

        "Finalise the module."

        for obj in self.all_objects:
            if isinstance(obj, (Class, Function)):
                obj.finalise_attributes()

    def add_object(self, obj, any_scope=0):

        """
        Record 'obj' if non-local or if the optional 'any_scope' is set to a
        true value.
        """

        if any_scope or not (self.namespaces and isinstance(self.namespaces[-1], Function)):
            self.all_objects.add(obj)

    # Optimisation tests.

    def should_optimise_unused_objects(self):
        return "unused_objects" in self.optimisations

    # Namespace methods.

    def store(self, name, obj):

        "Record attribute or local 'name', storing 'obj'."

        if not self.namespaces:
            self.set(name, obj, not self.in_loop)
        else:
            self.namespaces[-1].set(name, obj, not self.in_loop)

    def store_lambda(self, obj):

        "Store a lambda function 'obj'."

        self.add_object(obj)

    def store_module_attr(self, name, module):

        """
        Record module attribute 'name' in the given 'module' using the current
        expression.
        """

        if isinstance(self.expr, Attr):
            assigned_value = self.expr.value
        else:
            assigned_value = self.expr

        module.set(name, assigned_value, 0)

    def store_class_attr(self, name):

        """
        Record class attribute 'name' in the current class using the current
        expression.
        """

        if self.in_method and self.namespaces[-2].has_key(name):

            if isinstance(self.expr, Attr):
                assigned_value = self.expr.value
            else:
                assigned_value = self.expr

            self.namespaces[-2].set(name, assigned_value, 0)
            return 1

        return 0

    def store_instance_attr(self, name):

        "Record instance attribute 'name' in the current class."

        if self.in_method:

            # Current namespace is the function.
            # Previous namespace is the class.

            self.namespaces[-2].add_instance_attribute(name)

    def get_parent(self):

        "Return the parent (or most recent) namespace currently exposed."

        return (self.namespaces[-1:] or [self])[0]

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError(self.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def NOP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return None

    def OP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return Instance()

    def _visitUnary(self, node):

        "Accounting method for the unary operator 'node'."

        method = unary_methods[node.__class__.__name__]
        self.importer.use_name(method)
        return self.OP(node)

    def _visitBinary(self, node):

        "Accounting method for the binary operator 'node'."

        left_method, right_method = binary_methods[node.__class__.__name__]
        self.importer.use_name(left_method)
        self.importer.use_name(right_method)
        return self.OP(node)

    def _visitFunction(self, node, name):

        """
        Return a function object for the function defined by 'node' with the
        given 'name'. If a lambda expression is being visited, 'name' should be
        None.
        """

        function = Function(
            name,
            self.get_parent(),
            node.argnames,
            node.defaults,
            (node.flags & 4 != 0),
            (node.flags & 8 != 0),
            self,
            node
            )

        # Make a back reference from the node for code generation.

        node.unit = function

        # Process the defaults.

        for n in node.defaults:
            self.expr = self.dispatch(n)
            if isinstance(self.expr, Attr):
                function.store_default(self.expr.value)
            else:
                function.store_default(self.expr)

        # Enter the function.

        self.namespaces.append(function)

        # Current namespace is the function.
        # Previous namespace is the class.

        if len(self.namespaces) > 1 and isinstance(self.namespaces[-2], Class):
            if name == "__init__":
                self.in_init = 1
            self.in_method = 1

        self.dispatch(node.code)
        self.in_init = 0
        self.in_method = 0
        self.namespaces.pop()

        if name is not None:
            self.store(name, function)
        else:
            self.store_lambda(function)

            # Lambda functions are always assumed to be referenced. This is
            # because other means of discovering the referencing of objects rely
            # on the original names inherent in the definition of those objects. 

            function.set_referenced()

            # Where defaults exist, an instance needs creating. Thus, it makes
            # no sense to return a reference to the function here, since the
            # recipient will not be referencing the function itself.

            if node.defaults:
                return Instance() # indicates no known target

        self.add_object(function, any_scope=1)
        return function

    # Specific handler methods.

    visitAdd = _visitBinary

    visitAnd = OP

    visitAssert = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if isinstance(expr, Attr):
            if expr.name == "self":
                if not self.store_class_attr(node.attrname):
                    self.store_instance_attr(node.attrname)
            elif isinstance(expr.value, Module):
                self.store_module_attr(node.attrname, expr.value)
                print "Warning: attribute %r of module %r set outside the module." % (node.attrname, expr.value.name)
        return None

    def visitAssList(self, node):
        for i, n in enumerate(node.nodes):
            self.dispatch(n)
            self.importer.make_constant(i) # for __getitem__(i) at run-time
        return None

    def visitAssName(self, node):
        self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    def visitAugAssign(self, node):

        # Accounting.

        aug_method, (left_method, right_method) = augassign_methods[node.op]
        self.importer.use_name(aug_method)
        self.importer.use_name(left_method)
        self.importer.use_name(right_method)
        return self.OP(node)

    visitBackquote = OP

    visitBitand = _visitBinary

    visitBitor = _visitBinary

    visitBitxor = _visitBinary

    visitBreak = NOP

    visitCallFunc = OP

    def visitClass(self, node):

        """
        Register the class at the given 'node' subject to the restrictions
        mentioned in the module docstring.
        """

        if self.namespaces:
            print "Class %r in %r is not global: ignored." % (node.name, self.namespaces[-1].full_name())
            return None
        else:
            cls = Class(node.name, self.get_parent(), self, node)

            # Visit the base class expressions, attempting to find concrete
            # definitions of classes.

            for base in node.bases:
                expr = self.dispatch(base)
                if isinstance(expr, Attr):
                    if expr.assignments != 1:
                        raise InspectError(self.full_name(), node,
                            "Base class %r for %r is not constant." % (base, cls.full_name()))
                    else:
                        cls.add_base(expr.value)
                else: # if expr is None:
                    raise InspectError(self.full_name(), node,
                        "Base class %r for %r is not found: it may be hidden in some way." % (base, cls.full_name()))

            # NOTE: Potentially dubious measure to permit __init__ availability.
            # If no bases exist, adopt the 'object' class.

            if not node.bases and not (self.name == "__builtins__" and node.name == "object") :
                expr = self.dispatch(compiler.ast.Name("object"))
                cls.add_base(expr.value)

            # Make a back reference from the node for code generation.

            node.unit = cls

            # Make an entry for the class.

            self.store(node.name, cls)
            self.add_object(cls)

            # Process the class body.

            self.namespaces.append(cls)
            self.dispatch(node.code)
            self.namespaces.pop()

            return cls

    def visitCompare(self, node):

        # Accounting.
        # NOTE: Replicates some code in micropython.ast.visitCompare.

        for op in node.ops:
            op_name, next_node = op
            methods = comparison_methods[op_name]
            if methods is not None:
                self.importer.use_name(methods[0])
                self.importer.use_name(methods[1])
            elif op_name.endswith("in"):
                self.importer.use_name("__contains__")

        return self.OP(node)

    def visitConst(self, node):
        return self.importer.make_constant(node.value)

    visitContinue = NOP

    visitDecorators = NOP

    visitDict = OP

    visitDiscard = NOP

    visitDiv = _visitBinary

    visitEllipsis = NOP

    visitExec = NOP

    visitExpression = OP

    visitFloorDiv = _visitBinary

    def visitFor(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0
        return None

    def visitFrom(self, node):
        module = self.importer.load(node.modname, 1)

        if module is not None:
            module.set_referenced()

        #if module is None:
        #    print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None and module.namespace.has_key(name):
                    attr = module[name]
                    self.store(alias or name, attr.value)
                    if isinstance(attr.value, Module) and not attr.value.loaded:
                        self.importer.load(attr.value.name)

                # Support the import of names from missing modules.

                else:
                    self.store(alias or name, UnresolvedName(name, node.modname, self))
            else:
                if module is not None:
                    for n in module.namespace.keys():
                        attr = module[n]
                        self.store(n, attr.value)
                        if isinstance(attr.value, Module) and not attr.value.loaded:
                            self.importer.load(attr.value.name)

        return None

    def visitFunction(self, node):
        return self._visitFunction(node, node.name)

    visitGenExpr = OP

    visitGenExprFor = NOP

    visitGenExprIf = NOP

    visitGenExprInner = NOP

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        attrname = node.attrname

        if isinstance(expr, Attr):
            value = expr.value
            if isinstance(value, (Class, Module)):
                attr = value.namespace.get(attrname)
            elif isinstance(value, UnresolvedName):
                attr = UnresolvedName(attrname, value.full_name(), self)
            else:
                attr = None
        elif self.builtins is not None:
            attr = self.builtins.get(attrname)
        else:
            attr = UnresolvedName(attrname, value.full_name(), self)

        # Accounting.

        if attr is not None:
            attr.set_referenced()

        self.importer.use_name(attrname)

        return attr

    def visitGlobal(self, node):
        if self.namespaces:
            for name in node.names:
                self.namespaces[-1].make_global(name)

                # Record a global entry for the name in the module.

                if not self.has_key(name):
                    self[name] = Global()

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitIfExp = NOP

    def visitImport(self, node):
        for name, alias in node.names:
            if alias is not None:
                module = self.importer.load(name, 1) or UnresolvedName(None, name, self)
                self.store(alias, module)
            else:
                module = self.importer.load(name) or UnresolvedName(None, name.split(".")[0], self)
                self.store(name.split(".")[0], module)
            module.set_referenced()

        return None

    visitInvert = _visitUnary

    def visitKeyword(self, node):
        self.dispatch(node.expr)
        self.importer.make_constant(node.name)
        self.keyword_names.add(node.name)
        return None

    def visitLambda(self, node):
        return self._visitFunction(node, None)

    visitLeftShift = _visitBinary

    visitList = OP

    visitListComp = OP

    visitListCompFor = NOP

    visitListCompIf = NOP

    visitMod = _visitBinary

    def visitModule(self, node):

        # Make a back reference from the node for code generation.

        node.unit = self
        return self.dispatch(node.node)

    visitMul = _visitBinary

    def visitName(self, node):
        name = node.name

        if self.importer.predefined_constants.has_key(name):
            attr = self.importer.get_predefined_constant(name)
        elif self.namespaces and self.namespaces[-1].has_key(name):
            attr = self.namespaces[-1][name]
        elif self.has_key(name):
            attr = self[name]
        elif self.builtins is not None and self.builtins.has_key(name):
            attr = self.builtins[name]
        else:
            attr = None

        # Accounting.

        if attr is not None:
            attr.set_referenced()

        self.importer.use_name(name)

        return attr

    visitNot = OP

    visitOr = OP

    visitPass = NOP

    visitPower = _visitBinary

    visitPrint = NOP

    visitPrintnl = NOP

    visitRaise = NOP

    visitReturn = NOP

    visitRightShift = _visitBinary

    visitSlice = OP

    visitSliceobj = OP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSub = _visitBinary

    visitSubscript = OP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:

            # Establish the local for the handler.

            if var is not None:
                self.dispatch(var)
            if n is not None:
                self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitTryFinally = NOP

    visitTuple = OP

    visitUnaryAdd = _visitUnary

    visitUnarySub = _visitUnary

    def visitWhile(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0
        return None

    visitWith = NOP

    visitYield = NOP

class Global:

    """
    A reference to an object assigned to a global from outside the module
    top-level.
    """

    pass

# vim: tabstop=4 expandtab shiftwidth=4
